"""Contains econet connection shortcut."""

from .econet import EcoNET
from .version import __version__  # noqa


def econet_connection(host: str, port: int, **kwargs):
    """Creates ecoNET connection instance.

    Keyword arguments:
    host -- hostname or ip of rs485 to tcp bridge,
        connected to ecoMAX controller
    port -- port of rs485 to tcp bridge,
        connected to ecoMAX controller
    **kwargs -- keyword arguments directly passed to asyncio's
        create_connection method
    """
    return EcoNET(host, port, **kwargs)



"""Contains constants."""

FRAME_START: int = 0x68
FRAME_END: int = 0x16
HEADER_SIZE: int = 7
BROADCAST_ADDRESS: int = 0x00
ECONET_ADDRESS: int = 0x56
ECONET_TYPE: int = 0x30
ECONET_VERSION: int = 0x05

READER_BUFFER_SIZE: int = 1000
READER_TIMEOUT: int = 5
RECONNECT_TIMEOUT: int = 30

UID_BASE: int = 32
UID_BASE_BITS: int = 5
UID_CHAR_BITS: int = 8

WLAN_ENCRYPTION_UNKNOWN = 0
WLAN_ENCRYPTION_NONE = 1
WLAN_ENCRYPTION_WEP = 2
WLAN_ENCRYPTION_WPA = 3
WLAN_ENCRYPTION_WPA2 = 4
WLAN_ENCRYPTION = (
    WLAN_ENCRYPTION_UNKNOWN,
    WLAN_ENCRYPTION_NONE,
    WLAN_ENCRYPTION_WEP,
    WLAN_ENCRYPTION_WPA,
    WLAN_ENCRYPTION_WPA2,
)

DEFAULT_IP = "0.0.0.0"
DEFAULT_NETMASK = "255.255.255.0"

MODULE_PANEL: str = "MODULE_PANEL"
MODULE_A: str = "MODULE_A"
MODULE_B: str = "MODULE_B"
MODULE_C: str = "MODULE_C"
MODULE_LAMBDA: str = "MODULE_LAMBDA"
MODULE_ECOSTER: str = "MODULE_ECOSTER"
MODULES: list = (
    MODULE_PANEL,
    MODULE_A,
    MODULE_B,
    MODULE_C,
    MODULE_LAMBDA,
    MODULE_ECOSTER,
)

MODES: list = (
    "Off",
    "Starting",
    "Kindling",
    "Heating",
    "Supervision",
    "Cooling",
    "Standby",
)

MIXER_TEMP: str = "TEMP"
MIXER_TARGET: str = "TARGET"
MIXER_PUMP: str = "PUMP"

ECOSTER_CONTACTS: str = "CONTACTS"
ECOSTER_SCHEDULE: str = "SCHEDULE"
ECOSTER_MODE: str = "MODE"
ECOSTER_TEMP: str = "TEMP"
ECOSTER_TARGET: str = "TARGET"

DATA_ALARMS: str = "ALARMS"
DATA_CO_STATUS: str = "CO_STATUS"
DATA_CO_TARGET: str = "CO_TARGET"
DATA_CWU_STATUS: str = "CWU_STATUS"
DATA_CWU_TARGET: str = "CWU_TARGET"
DATA_FAN_POWER: str = "FAN_POWER"
DATA_FRAMES: str = "FRAMES"
DATA_FUEL_CONSUMPTION: str = "FUEL_CONSUMPTION"
DATA_FUEL_LEVEL: str = "FUEL_LEVEL"
DATA_LAMBDA_LEVEL: str = "LAMBDA_LEVEL"
DATA_LAMBDA_STATUS: str = "LAMBDA_STATUS"
DATA_LAMBDA_TARGET: str = "LAMBDA_TARGET"
DATA_MIXERS: str = "MIXERS"
DATA_MODE: str = "MODE"
DATA_POWER: str = "POWER"
DATA_LOAD: str = "LOAD"
DATA_THERMOSTAT: str = "THERMOSTAT"
DATA_THERMOSTATS: str = "THERMOSTATS"
DATA_TRANSMISSION: str = "TRANSMISSION"

TEMPERATURES: list = (
    "CO_TEMP",
    "FEEDER_TEMP",
    "CWU_TEMP",
    "OUTSIDE_TEMP",
    "BACK_TEMP",
    "EXHAUST_TEMP",
    "OPTICAL_TEMP",
    "UPPER_BUFFER_TEMP",
    "LOWER_BUFFER_TEMP",
    "UPPER_SOLAR_TEMP",
    "LOWER_SOLAR_TEMP",
    "FIREPLACE_TEMP",
    "TOTAL_GAIN",
    "HYDRAULIC_COUPLER_TEMP",
    "EXCHANGER_TEMP",
    "AIR_IN_TEMP",
    "AIR_OUT_TEMP",
)

OUTPUTS: list = (
    "FAN",
    "FEEDER",
    "CO_PUMP",
    "CWU_PUMP",
    "CICULATION_PUMP",
    "LIGHTER",
    "ALARM",
    "OUTER_BOILER",
    "FAN2_EXHAUST",
    "FEEDER2",
    "OUTER_FEEDER",
    "SOLAR_PUMP",
    "FIREPLACE_PUMP",
    "GCZ_CONTACT",
    "BLOW_FAN1",
    "BLOW_FAN2",
)

DATA_CO_PUMP_FLAG: str = "CO_PUMP_FLAG"
DATA_CWU_PUMP_FLAG: str = "CWU_PUMP_FLAG"
DATA_CIRCULATION_PUMP_FLAG: str = "CIRCULATION_PUMP_FLAG"
DATA_SOLAR_PUMP_FLAG: str = "SOLAR_PUMP_FLAG"
FLAGS: list = (
    DATA_CO_PUMP_FLAG,
    DATA_CWU_PUMP_FLAG,
    DATA_CIRCULATION_PUMP_FLAG,
    DATA_SOLAR_PUMP_FLAG,
)

STATUSES: list = (
    "CO_TARGET",
    "CO_STATUS",
    "CWU_TARGET",
    "CWU_STATUS",
)

CURRENT_DATA: list = [
    DATA_ALARMS,
    DATA_FAN_POWER,
    DATA_FRAMES,
    DATA_FUEL_CONSUMPTION,
    DATA_FUEL_LEVEL,
    DATA_LAMBDA_LEVEL,
    DATA_LAMBDA_STATUS,
    DATA_LAMBDA_TARGET,
    DATA_LOAD,
    DATA_MIXERS,
    DATA_MODE,
    DATA_POWER,
    DATA_THERMOSTAT,
    DATA_THERMOSTATS,
    DATA_TRANSMISSION,
]
CURRENT_DATA.extend(TEMPERATURES)
CURRENT_DATA.extend(OUTPUTS)
CURRENT_DATA.extend(FLAGS)
CURRENT_DATA.extend(STATUSES)
CURRENT_DATA.extend(MODULES)

EDITABLE_PARAMS: list = (
    "AIRFLOW_POWER_100",
    "AIRFLOW_POWER_50",
    "AIRFLOW_POWER_30",
    "POWER_100",
    "POWER_50",
    "POWER_30",
    "MAX_FAN_BOILER_POWER",
    "MIN_FAN_BOILER_POWER",
    "FUEL_FEEDING_TIME_100",
    "FUEL_FEEDING_TIME_50",
    "FUEL_FEEDING_TIME_30",
    "FUEL_FEEDING_BREAK_100",
    "FUEL_FEEDING_BREAK_50",
    "FUEL_FEEDING_BREAK_30",
    "CYCLE_TIME",
    "H2_HYSTERESIS",
    "H1_HYSTERESIS",
    "BOILER_HYSTERESIS",
    "CONTROL_MODE",
    "MIN_FL_POWER",
    "MAX_FL_POWER",
    "MIN_BOILER_POWER",
    "MAX_BOILER_POWER",
    "MIN_FAN_POWER",
    "MAX_FAN_POWER",
    "T_REDUCTION_AIRFLOW",
    "FAN_POWER_GAIN",
    "FUEL_FLOW_CORRECTION_FL",
    "FUEL_FLOW_CORRECTION",
    "AIRFLOW_CORRECTION_100",
    "FEEDER_CORRECTION_100",
    "AIRFLOW_CORRECTION_50",
    "FEEDER_CORRECTION_50",
    "AIRFLOW_CORRECTION_30",
    "FEEDER_CORRECTION_30",
    "AIRFLOW_POWER_GRATE",
    "HIST_BOILER_GRATE",
    "SUPERVISION_WORK_AIRFLOW",
    "SUPERVISION_WORK_AIRFLOW_BRAKE",
    "CO_TEMP_GRATE",
    "DET_TIME_FUEL_GRATE",
    "AIRFLOW_POWER_KINDLE",
    "SMALL_AIRFLOW_POWER_KINDLE",
    "AIRFLOW_KINDLE_DELAY",
    "SCAVENGE_KINDLE",
    "FEEDER_KINDLE",
    "FEEDER_KINDLE_WEIGHT",
    "KINDLE_TIME",
    "WARMING_UP_TIME",
    "FUMES_TEMP_KINDLE_FINISH",
    "FINISH_KINDLE_THRESHOLD",
    "FUMES_DELTA_KINDLE",
    "DELTA_T_KINDLE",
    "MIN_KINDLE_POWER_TIME",
    "SCAVENGE_AFTER_KINDLE",
    "AIRFLOW_POWER_AFTER_KINDLE",
    "SUPERVISION_TIME",
    "FEED_TIME_SUPERVISION",
    "FEED_TIME_SUPERVISION_WEIGHT",
    "FEED_SUPERVISION_BREAK",
    "SUPERVISION_CYCLE_DURATION",
    "AIRFLOW_POWER_SUPERVISION",
    "FAN_SUPERVISON_BREAK",
    "FAN_WORK_SUPERVISION",
    "INCREASE_FAN_SUPPORT_MODE",
    "MAX_EXTINGUISH_TIME",
    "MIN_EXTINGUISH_TIME",
    "EXTINGUISH_TIME",
    "AIRFLOW_POWER_EXTINGUISH",
    "AIRFLOW_WORK_EXTINGUISH",
    "AIRFLOW_BRAKE_EXTINGUISH",
    "SCAVENGE_START_EXTINGUISH",
    "SCAVENGE_STOP_EXTINGUISH",
    "CLEAN_BEGIN_TIME",
    "EXTINGUISH_CLEAN_TIME",
    "AIRFLOW_POWER_CLEAN",
    "WARMING_UP_BRAKE_TIME",
    "WARMING_UP_CYCLE_TIME",
    "REMIND_TIME",
    "LAMBDA_WORK",
    "LAMBDA_CORRECTION_RANGE",
    "OXYGEN_100",
    "OXYGEN_50",
    "OXYGEN_30",
    "OXYGEN_CORRECTION_FL",
    "FUEL_KG_H",
    "FEEDER_CALIBRATION",
    "FUEL_FACTOR",
    "CALORIFIC_KWH_KG",
    "FUEL_DETECTION_TIME",
    "FUMES_TEMP_FUEL_DETECTION",
    "SCHEDULE_FEEDER_2",
    "FEED2_H1",
    "FEED2_H2",
    "FEED2_H3",
    "FEED2_H4",
    "FEED2_WORK",
    "FEED2_BREAK",
    "CO_TEMP_SET",
    "MIN_SET_CO_TEMP",
    "MAX_SET_CO_TEMP",
    "SWITCH_CO_TEMP",
    "PAUSE_CO_CWU",
    "PAUSE_TERM",
    "WORK_TERM",
    "INCREASE_TEMP_CO",
    "PROGRAM_CONTROL_CO",
    "CO_HEAT_CURVE",
    "PARALLEL_CO_HEAT_CURVE",
    "WEATHER_FACTOR",
    "TERM_BOILER_OPERATION",
    "TERM_BOILER_MODE",
    "DECREASE_SET_CO_TERM",
    "TERM_PUMP_OFF",
    "AL_BOILER_TEMP",
    "MAX_FEED_TEMP",
    "EXTERN_BOILER_TEMP",
    "ALARM_NOTIF",
    "PUMP_HYSTERESIS",
    "CWU_SET_TEMP",
    "MIN_CWU_SET_TEMP",
    "MAX_CWU_SET_TEMP",
    "CWU_WORK_MODE",
    "CWU_HYSTERESIS",
    "CWU_DISINFECTION",
    "AUTO_SUMMER",
    "SUMMER_TEMP_ON",
    "SUMMER_TEMP_OFF",
    "CWU_FEEDING_EXTENSION",
    "CIRCULATION_CONTROL",
    "CIRCULATION_PAUSE_TIME",
    "CIRCULATION_WORK_TIME",
    "CIRCULATION_START_TEMP",
    "BUFFER_CONTROL",
    "BUFFER_MAX_TEMP",
    "MIN_BUFFER_TEMP",
    "BUFFER_HISTERESIS",
    "BUFFER_LOAD_START",
    "BUFFER_LOAD_STOP",
    "BOILER_CONTROL",
)

MIXERS_PARAMS: list = (
    "MIX_SET_TEMP",
    "MIN_MIX_SET_TEMP",
    "MAX_MIX_SET_TEMP",
    "LOW_MIX_SET_TEMP",
    "CTRL_WEATHER_MIX",
    "MIX_HEAT_CURVE",
    "PARALLEL_OFFSET_HEAT_CURV",
    "WEATHER_TEMP_FACTOR",
    "MIX_OPERATION",
    "MIX_INSENSITIVITY",
    "MIX_THERM_OPERATION",
    "MIX_THERM_MODE",
    "MIX_OFF_THERM_PUMP",
    "MIX_SUMMER_WORK",
)



"""Contains classes for supported devices."""
from __future__ import annotations

from .constants import (
    CURRENT_DATA,
    DATA_FRAMES,
    DATA_MODE,
    EDITABLE_PARAMS,
    MODES,
    MODULE_PANEL,
)
from .frame import Request
from .frames.requests import BoilerControl, SetParameter
from .storage import FrameBucket

ECOMAX_ADDRESS: int = 0x45
ECOSTER_ADDRESS: int = 0x51


class Device:
    """A device representation.

    Passed to the user-defined callback method.
    """

    def __init__(self, data: dict = None, parameters: dict = None):
        """Creates device.

        Keyword arguments:
        data -- device data
        parameters -- editable parameters
        """

        self.__dict__["_data"] = {}
        self.__dict__["_parameters"] = {}

        if data is not None:
            self.set_data(data)

        if parameters is not None:
            self.set_parameters(parameters)

        self.product = None
        self.uid = None
        self.password = None
        self.struct = []
        self.bucket = FrameBucket()
        self._queue = []
        self._is_on = False

    def __getattr__(self, name: str) -> None:
        """Gets current data item as class attribute.

        Keyword arguments:
        name -- name of property to get
        """
        key = name.upper()
        if key in self._data:
            return self._data[key]

        if key in self._parameters:
            return self._parameters[key]

        return None

    def __setattr__(self, name: str, value) -> None:
        """Sets class attribute or device parameter.

        Keyword arguments:
        name -- attribute name
        value -- attribute value
        """
        key = name.upper()
        if key in self._data:
            raise AttributeError()

        if key in self._parameters:
            self._parameters[key].set(value)
            self._queue.append(self._parameters[key].request)
        else:
            self.__dict__[name] = value

    def has_data(self) -> bool:
        """Checks if device instance has any data."""
        return bool(self._data)

    def set_data(self, data: dict) -> None:
        """Sets device data received in CurrentData frame.

        Keyword arguments:
        data - data parsed from CurrentData response frame
        """
        self.bucket.fill(data[DATA_FRAMES])
        for name, value in data.items():
            if name in CURRENT_DATA:
                self._data[name] = value

    def has_parameters(self) -> bool:
        """Check if device instance has parameters."""
        return bool(self._parameters)

    def set_parameters(self, parameters: dict) -> None:
        """Sets device settings received in parameters frame."""
        for name, parameter in parameters.items():
            if name in EDITABLE_PARAMS:
                self._parameters[name] = Parameter(
                    name=name,
                    value=parameter["value"],
                    min_=parameter["min"],
                    max_=parameter["max"],
                )

        self._parameters["BOILER_CONTROL"] = Parameter(
            name="BOILER_CONTROL", value=int(self.is_on), min_=0, max_=1
        )

    @property
    def software(self) -> str:
        """Returns software version."""
        try:
            return self._data[MODULE_PANEL]
        except KeyError:
            return None

    @property
    def is_on(self) -> bool:
        """Returns current state."""
        try:
            return bool(self._data[DATA_MODE] != 0)
        except KeyError:
            return False

    @property
    def mode(self) -> str:
        """Returns current mode."""
        if self._data:
            mode = self._data[DATA_MODE]
            try:
                return MODES[mode]
            except IndexError:
                pass

        return "Unknown"

    @property
    def queue(self):
        """Clears and returns changed parameters queue."""
        queue = self._queue
        self._queue = []
        return queue

    @property
    def data(self):
        """Returns device data."""
        return self._data

    @property
    def parameters(self):
        """Returns device parameters."""
        return self._parameters

    @property
    def changes(self) -> list[Parameter]:
        """Returns changed device parameters."""
        changes = self.queue
        changes.extend(self.bucket.queue)
        return changes

    def __str__(self) -> str:
        """Converts device instance to a string."""
        output = f"""
Product:        {self.product}
Software Ver.:  {self.software}
UID:            {self.uid}
Password:       {self.password}
"""

        if self.has_data():
            output += "\nCurrent data:\n"
            for k, v in self._data.items():
                output += f" -- {k}: {v}\n"

        if self.has_parameters():
            output += "\nEditable parameters:\n"
            for _, parameter in self._parameters.items():
                output += f" -- {parameter}\n"

        return output.lstrip()


class EcoMAX(Device):
    """ecoMAX device representation."""

    address = ECOMAX_ADDRESS


class EcoSTER(Device):
    """ecoSTER device representation."""

    address = ECOSTER_ADDRESS


class DeviceCollection:
    """Collection of ecoNET devices."""

    def __init__(self):
        """Creates device collection."""
        self._classes = Device.__subclasses__()
        self._addresses = [v.address for v in self._classes]
        self._instances = {}
        self._names = {}

    def __getattr__(self, name: str) -> Device | None:
        """Gets device by name.""

        Keyword arguments:
        name -- name of device to get
        """
        name = name.lower()
        if name in self._names:
            return self._instances[self._names[name]]

        return None

    def has(self, needle) -> bool:
        """Checks if collection has device for specified address.

        Keyword arguments:
        needle -- address or name of device
        """
        return (needle in self._addresses) or (needle in self._names)

    def get(self, address: int) -> Device | None:
        """Gets device by address.

        Keyword arguments:
        address -- int address of device
        """
        if address not in self._instances:
            try:
                index = self._addresses.index(address)
                cls = self._classes[index]
            except KeyError:
                return None

            self._instances[address] = cls()
            self._names[cls.__name__.lower()] = address

        return self._instances[address]


class Parameter:
    """Device parameter representation."""

    def __init__(self, name: str, value, min_: int, max_: int):
        """Creates parameter."""
        self.name = name
        self.value = int(value)
        self.min_ = min_
        self.max_ = max_

    def set(self, value) -> None:
        """Sets parameter value.

        Keyword arguments:
        value -- new value to set parameter to
        """
        if self.value != value and self.min_ <= value <= self.max_:
            self.value = value

    @property
    def request(self) -> Request:
        """Returns request to change parameter."""
        if self.name == "BOILER_CONTROL":
            return BoilerControl(data=self.__dict__)

        return SetParameter(data=self.__dict__)

    def __repr__(self) -> str:
        """Returns serializable string representation."""
        return f"""Parameter(
    name = {self.name},
    value = {self.value},
    min_ = {self.min_},
    max_ = {self.max_}
)""".strip()

    def __str__(self) -> str:
        """Returns string representation."""
        return f"{self.name}: {self.value} (range {self.min_} - {self.max_})"

    def __int__(self) -> int:
        """Returns integer representation of parameter value.

        Keyword arguments:
        other -- other value to compare to
        """
        return int(self.value)

    def __eq__(self, other) -> int:
        """Compares if parameter value is equal to other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value == other

    def __ge__(self, other) -> int:
        """Compares if parameter value is greater or equal to other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value >= other

    def __gt__(self, other) -> int:
        """Compares if parameter value is greater than other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value > other

    def __le__(self, other) -> int:
        """Compares if parameter value is less or equal to other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value <= other

    def __lt__(self, other) -> int:
        """Compares if parameter value is less that other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value < other
